# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import uuid
import logging
import celery
import json
import yaml

from django.core.files import File
from django.utils import timezone

from . import models, utils

logger = logging.getLogger(__name__)


class Task(celery.Task):
    @property
    def task(self):
        if not hasattr(self, '_task') or self._task is None:
            if self.request.id is None:
                task_id = uuid.uuid4().hex
            else:
                task_id = self.request.id
            self._task, _ = models.CeleryTask.objects.get_or_create(task_id=task_id)
        return self._task

    def apply_async(self, args=None, kwargs=None, *args_, **kwargs_):
        user = kwargs_.get('user')
        if user is None:
            user = utils.get_user()
        result = super(Task, self).apply_async(args=args, kwargs=kwargs, *args_, **kwargs_)
        logger.debug('created celery task : {} {} name={}'.format(user, result.task_id, self.name))
        models.CeleryTask.objects.create(
            user=user, task_id=result.task_id, task_path=self.name,
            args=yaml.dump(args if args is not None else ()),
            kwargs=yaml.dump(kwargs if kwargs is not None else {}))
        return result

    def __call__(self, *args, **kwargs):
        self._task = None
        logger.debug("{} {} is started.".format(self.request.id, self.task))
        try:
            utils.set_user(self.task.user)
            if self.task.status == 'cancel':
                self.task.status = 'cancelled'
                return
            elif self.task.status != 'sent':
                self.task.status = 'retry_rejected'
                return
            self.task.status = 'started'
            self.task.start_at = timezone.now()
            self.task.save()
            return self.run(*args, **kwargs)
        except SystemExit:
            raise  # no log in revoke
        except:
            logger.exception('celery task failed')
            raise
        finally:
            self.task.end_at = timezone.now()
            self.task.save()
            utils.reset_user()
            logger.debug("{} is terminated.".format(self.request.id))

    def on_success(self, retval, task_id, args, kwargs):
        if (self.task.status == 'revoked' or self.task.status == 'cancelled' or
                self.task.status == 'retry_rejected'):
            return
        self.task.status = 'succeeded'
        self.task.save()
        logger.debug("{} is on success.".format(self.request.id))

    def on_retry(self, exc, task_id, args, kwargs, einfo):
        self.task.status = 'retried'
        self.task.stack_trace = str(einfo)
        self.task.save()
        logger.debug("{} is on retry. {}".format(self.request.id, einfo))

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        self.task.status = 'failed'
        self.task.stack_trace = str(einfo)
        self.task.save()
        logger.debug("{} is on failure. {}".format(self.request.id, einfo))

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        if self.task.status == 'succeeded':
            self.task.result_string = json.dumps(retval)
            self.task.save()
        logger.debug("{} after_return status={}.".format(self.request.id, status))


class FileTask_(Task):

    def __call__(self, *args, **kwargs):
        retval = super(FileTask_, self).__call__(*args, **kwargs)
        if self.task.status == 'succeeded':
            filename = '{}.{}'.format(uuid.uuid4().hex, self.ext)
            if hasattr(retval, 'read'):
                self.task.result_file.save(filename, File(retval), save=True)
                self.task.save()
            else:
                return None
            return filename


def FileTask(extension, base_class=FileTask_):
    name = extension[0].upper() + extension[1].lower()
    return type('{}FileTask'.format(name), (base_class, ), {'ext': extension})



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-10 06:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('heavy_celery', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('worker_id', models.CharField(max_length=256, verbose_name='\u30ef\u30fc\u30ab\u30fcID')),
                ('started_at', models.DateTimeField(auto_now_add=True)),
                ('ended_at', models.DateTimeField(blank=True, null=True)),
                ('beated_at', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(choices=[(0, '\u958b\u59cb'), (1, '\u505c\u6b62'), (2, '\u5fdc\u7b54\u306a\u3057')], default=0)),
            ],
        ),
        migrations.CreateModel(
            name='WorkerTaskLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.CharField(max_length=256, unique=True)),
                ('task_path', models.CharField(max_length=256)),
                ('started_at', models.DateTimeField(auto_now_add=True)),
                ('ended_at', models.DateTimeField(blank=True, null=True)),
                ('worker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='heavy_celery.Worker')),
            ],
        ),
    ]



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import logging

from django.db.models import Q, F
from . import models, utils

logger = logging.getLogger(__name__)


def spawner():
    '''
    from <project_name>.apps.heavy_celery.cron import spawner as _cron_scheduler
    @app.task()
    def cron_scheduler():
        _cron_scheduler()


    app.conf.beat_schedule = {
        'cron_scheduler': {
            'task': '<project_name>.apps.common.tasks.cron_scheduler',
            'schedule': crontab(),
            'args': (),
            'options': dict(queue='time_sensitive', routing_key='time_sensitive_tasks'),
        },
    }
    '''
    now = utils.get_now()
    logger.debug('cron spawner {}'.format(now))
    for schedule in models.CronSchedule.objects.filter(
            Q(next_time__lte=now),
            Q(total_run_count__lt=F('max_run_count')) | Q(max_run_count=-1),
            Q(expires__gt=now) | Q(expires__isnull=True)).select_related('task'):
        logger.debug('cron spawner {} {} run start'.format(now, schedule))
        try:
            schedule.run(now)
        except:
            logger.exception('{} failed'.format(schedule))
        logger.debug('cron spawner {} {} run end'.format(now, schedule))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-10 00:52
from __future__ import unicode_literals

from django.conf import settings
import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.CharField(max_length=256, unique=True, verbose_name='\u30bf\u30b9\u30afID')),
                ('task_path', models.CharField(blank=True, max_length=256, null=True, verbose_name='\u30bf\u30b9\u30af\u540d')),
                ('args', models.TextField(blank=True, null=True, verbose_name='\u5f15\u6570')),
                ('kwargs', models.TextField(blank=True, null=True, verbose_name='kw\u5f15\u6570')),
                ('status', models.CharField(choices=[('sent', 'PENDING'), ('received', 'RECEIVED'), ('started', 'STARTED'), ('failed', 'FAILURE'), ('retried', 'RETRY'), ('succeeded', 'SUCCESS'), ('revoked', 'REVOKED'), ('rejected', 'REJECTED'), ('cancel', 'CANCEL'), ('cancelled', 'CANCELLED'), ('revoking', 'REVOKING'), ('retry_rejected', 'RETRY_REJECTED')], default='sent', max_length=64, verbose_name='\u30b9\u30c6\u30fc\u30bf\u30b9')),
                ('stack_trace', models.TextField(blank=True, null=True, verbose_name='\u30b9\u30bf\u30c3\u30af\u30c8\u30ec\u30fc\u30b9')),
                ('result_string', models.TextField(blank=True, null=True, verbose_name='\u7d50\u679c')),
                ('result_file', models.FileField(blank=True, null=True, storage=django.core.files.storage.FileSystemStorage(), upload_to=b'', verbose_name='\u7d50\u679c\u30d5\u30a1\u30a4\u30eb')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('start_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('end_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='CronSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=256, null=True, verbose_name='\u540d\u524d')),
                ('description', models.TextField(blank=True, null=True, verbose_name='\u8a73\u7d30')),
                ('cron_expr', models.CharField(max_length=256, verbose_name='cron')),
                ('next_time', models.DateTimeField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('expires', models.DateTimeField(blank=True, null=True, verbose_name='\u6709\u52b9\u671f\u9650')),
                ('last_run_at', models.DateTimeField(blank=True, null=True, verbose_name='\u6700\u7d42\u5b9f\u884c\u65e5\u6642')),
                ('total_run_count', models.PositiveIntegerField(default=0, editable=False, verbose_name='\u5b9f\u884c\u56de\u6570')),
                ('max_run_count', models.IntegerField(default=-1, verbose_name='\u6700\u5927\u7e70\u308a\u8fd4\u3057\u56de\u6570')),
            ],
        ),
        migrations.CreateModel(
            name='TaskSignature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=256, null=True, verbose_name='\u540d\u524d')),
                ('description', models.TextField(blank=True, null=True, verbose_name='\u8a73\u7d30')),
                ('task_path', models.CharField(max_length=256, verbose_name='\u30bf\u30b9\u30af\u540d')),
                ('args', models.TextField(blank=True, null=True, verbose_name='\u5f15\u6570')),
                ('kwargs', models.TextField(blank=True, null=True, verbose_name='kw\u5f15\u6570')),
                ('options', models.TextField(blank=True, null=True, verbose_name='\u30bf\u30b9\u30af\u30aa\u30d7\u30b7\u30e7\u30f3')),
            ],
        ),
        migrations.AddField(
            model_name='cronschedule',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='heavy_celery.TaskSignature'),
        ),
    ]



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

# import datetime
import uuid
import logging

# Task Signals
from celery.signals import before_task_publish
from celery.signals import after_task_publish
from celery.signals import task_prerun
from celery.signals import task_postrun
from celery.signals import task_retry
from celery.signals import task_success
from celery.signals import task_failure
from celery.signals import task_revoked
from celery.signals import task_unknown
from celery.signals import task_rejected

# App Signals
from celery.signals import import_modules

# Worker Signals
from celery.signals import celeryd_after_setup
from celery.signals import celeryd_init
from celery.signals import worker_init
from celery.signals import worker_ready
from celery.signals import heartbeat_sent
from celery.signals import worker_process_init
from celery.signals import worker_process_shutdown
from celery.signals import worker_shutdown

# Beat Signals
from celery.signals import beat_init
from celery.signals import beat_embedded_init

# Eventlet Signals
from celery.signals import eventlet_pool_started
from celery.signals import eventlet_pool_preshutdown
from celery.signals import eventlet_pool_postshutdown
from celery.signals import eventlet_pool_apply

# Logging Signals
from celery.signals import setup_logging
from celery.signals import after_setup_logger
from celery.signals import after_setup_task_logger

# Command signals
from celery.signals import user_preload_options

# Deprecated Signals
# from celery.signals import task_sent
from django.utils import timezone

from . import models, utils


logger = logging.getLogger(__name__)
_worker_id = None


def _get_worker_id():
    global _worker_id
    if _worker_id is None:
        _worker_id = uuid.uuid4().hex
    return _worker_id


@before_task_publish.connect
def before_task_publish_handler(*args, **kwargs):
    logger.debug('Task Signals before_task_publish {} {}'.format(args, kwargs))


@after_task_publish.connect
def after_task_publish_handler(*args, **kwargs):
    logger.debug('Task Signals after_task_publish {} {}'.format(args, kwargs))


@task_prerun.connect
def task_prerun_handler(*args, **kwargs):
    logger.debug('Task Signals task_prerun {} {}'.format(args, kwargs))
    worker, _ = models.Worker.objects.update_or_create(
        worker_id=_get_worker_id(), defaults=dict(beated_at=timezone.now()))
    models.WorkerTaskLog.objects.create(
        worker=worker, task_id=kwargs['task_id'], task_path=kwargs['task'].name)


@task_postrun.connect
def task_postrun_handler(*args, **kwargs):
    logger.debug('Task Signals task_postrun {} {}'.format(args, kwargs))
    t = models.WorkerTaskLog.objects.get(task_id=kwargs['task_id'])
    t.ended_at = timezone.now()
    t.save()


@task_retry.connect
def task_retry_handler(*args, **kwargs):
    logger.debug('Task Signals task_retry {} {}'.format(args, kwargs))


@task_success.connect
def task_success_handler(*args, **kwargs):
    logger.debug('Task Signals task_success {} {}'.format(args, kwargs))


@task_failure.connect
def task_failure_handler(*args, **kwargs):
    logger.debug('Task Signals task_failure {} {}'.format(args, kwargs))


@task_revoked.connect
def task_revoked_handler(request=None, *args, **kwargs):
    try:
        task = models.CeleryTask.objects.get(task_id=request.id)
        task.status = 'revoked'
        task.save()
    except models.CeleryTask.DoesNotExist:
        pass
    logger.debug('Task Signals task_revoked {} {}'.format(args, kwargs))
    utils.reset_user()


@task_unknown.connect
def task_unknown_handler(*args, **kwargs):
    logger.debug('Task Signals task_unknown {} {}'.format(args, kwargs))


@task_rejected.connect
def task_rejected_handler(*args, **kwargs):
    logger.debug('Task Signals task_rejected {} {}'.format(args, kwargs))


@import_modules.connect
def import_modules_handler(*args, **kwargs):
    logger.debug('App Signals import_modules {} {}'.format(args, kwargs))


@celeryd_after_setup.connect
def celeryd_after_setup_handler(*args, **kwargs):
    logger.debug('Worker Signals celeryd_after_setup {} {}'.format(args, kwargs))


@celeryd_init.connect
def celeryd_init_handler(*args, **kwargs):
    logger.debug('Worker Signals celeryd_init {} {}'.format(args, kwargs))


@worker_init.connect
def worker_init_handler(*args, **kwargs):
    logger.debug('Worker Signals worker_init {} {}'.format(args, kwargs))


@worker_ready.connect
def worker_ready_handler(*args, **kwargs):
    logger.debug('Worker Signals worker_ready {} {}'.format(args, kwargs))


@heartbeat_sent.connect
def heartbeat_sent_handler(*args, **kwargs):
    logger.debug('Worker Signals heartbeat_sent {} {}'.format(args, kwargs))
    # models.Worker.objects.update_or_create(
    #     worker_id=_get_worker_id(), defaults=dict(beated_at=timezone.now()))
    # timeout = timezone.now() - datetime.timedelta(minutes=10)
    # models.Worker.objects.filter(beated_at__lt=timeout).update(status=models.Worker.State.BEAT_FAILED)


@worker_process_init.connect
def worker_process_init_handler(*args, **kwargs):
    logger.debug('Worker Signals worker_process_init {} {}'.format(args, kwargs))
    # models.Worker.objects.get_or_create(worker_id=_get_worker_id())


@worker_process_shutdown.connect
def worker_process_shutdown_handler(*args, **kwargs):
    # worker, _ = models.Worker.objects.get_or_create(worker_id=_get_worker_id())
    # worker.ended_at = timezone.now()
    # worker.status = models.Worker.State.STOPPED
    # worker.save()

    logger.debug('Worker Signals worker_process_shutdown {} {}'.format(args, kwargs))
    with open('/data/static/django-celery.log', 'a') as fp:
        fp.writer('Worker Signals worker_process_shutdown {} {}\n'.format(args, kwargs))


@worker_shutdown.connect
def worker_shutdown_handler(*args, **kwargs):
    logger.debug('Worker Signals worker_shutdown {} {}'.format(args, kwargs))
    with open('/data/static/django-celery2.log', 'a') as fp:
        fp.writer('Worker Signals worker_shutdown {} {}\n'.format(args, kwargs))


@beat_init.connect
def beat_init_handler(*args, **kwargs):
    logger.debug('Beat Signals beat_init {} {}'.format(args, kwargs))


@beat_embedded_init.connect
def beat_embedded_init_handler(*args, **kwargs):
    logger.debug('Beat Signals beat_embedded_init {} {}'.format(args, kwargs))


@eventlet_pool_started.connect
def eventlet_pool_started_handler(*args, **kwargs):
    logger.debug('Eventlet Signals eventlet_pool_started {} {}'.format(args, kwargs))


@eventlet_pool_preshutdown.connect
def eventlet_pool_preshutdown_handler(*args, **kwargs):
    logger.debug('Eventlet Signals eventlet_pool_preshutdown {} {}'.format(args, kwargs))


@eventlet_pool_postshutdown.connect
def eventlet_pool_postshutdown_handler(*args, **kwargs):
    logger.debug('Eventlet Signals eventlet_pool_postshutdown {} {}'.format(args, kwargs))


@eventlet_pool_apply.connect
def eventlet_pool_apply_handler(*args, **kwargs):
    logger.debug('Eventlet Signals eventlet_pool_apply {} {}'.format(args, kwargs))


@setup_logging.connect
def setup_logging_handler(*args, **kwargs):
    logger.debug('Logging Signals setup_logging {} {}'.format(args, kwargs))


@after_setup_logger.connect
def after_setup_logger_handler(*args, **kwargs):
    logger.debug('Logging Signals after_setup_logger {} {}'.format(args, kwargs))


@after_setup_task_logger.connect
def after_setup_task_logger_handler(*args, **kwargs):
    logger.debug('Logging Signals after_setup_task_logger {} {}'.format(args, kwargs))


@user_preload_options.connect
def user_preload_options_handler(*args, **kwargs):
    logger.debug('Command signals user_preload_options {} {}'.format(args, kwargs))


# @task_sent.connect
# def task_sent_handler(*args, **kwargs):
#     logger.debug('Deprecated Signals task_sent {} {}'.format(args, kwargs))




# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import threading
from datetime import datetime

# from django.conf import settings
from django.utils import timezone
# from django.utils.module_loading import import_string
from croniter import croniter


_app = None
_user = {}
_middleware = None


def set_user(user):
    _user[threading.current_thread()] = user


def reset_user():
    _user.pop(threading.current_thread(), None)


def get_user():
    global _middleware

    user = _user.get(threading.current_thread(), None)
    if user:
        return user

    if _middleware is None:
        try:
            from crequest.middleware import CrequestMiddleware
        except ImportError:
            class CrequestMiddleware(object):
                @classmethod
                def get_request(klass):
                    return None
        _middleware = CrequestMiddleware

    request = _middleware.get_request()
    if request:
        return request.user
    return None


def get_now():
    return timezone.localtime(timezone.now()).replace(second=0, microsecond=0)


def get_next_cron(cron_expr, now=None):
    if now is None:
        now = get_now()
    return croniter(cron_expr, now).get_next(datetime)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from django.contrib import admin

from . import models


@admin.register(models.Worker)
class WorkerAdmin(admin.ModelAdmin):
    model = models.Worker
    list_display = ('id', 'status', 'worker_id', 'started_at', 'ended_at', 'beated_at')
    list_filter = ('status', )


@admin.register(models.WorkerTaskLog)
class WorkerTaskLogAdmin(admin.ModelAdmin):
    model = models.WorkerTaskLog
    list_display = ('id', 'worker', 'task_id', 'task_path', 'started_at', 'ended_at')


@admin.register(models.CronSchedule)
class CronScheduleAdmin(admin.ModelAdmin):
    model = models.CronSchedule
    list_display = ('id', 'name', 'cron_expr', 'task', 'next_time',
                    'expires', 'total_run_count', 'max_run_count')
    inline_reverse = ['task']


def execute_task_signature(modeladmin, request, queryset):
    for i in queryset:
        i.run()


execute_task_signature.short_description = "Taskを実行する"


@admin.register(models.TaskSignature)
class TaskSignatureAdmin(admin.ModelAdmin):
    model = models.TaskSignature
    list_display = ('id', 'name', 'task_path', 'args', 'kwargs', 'options')
    actions = [execute_task_signature, ]


def reexecute_tasks(modeladmin, request, queryset):
    for i in queryset:
        i.reexecute()


def cancel_revoke_tasks(modeladmin, request, queryset):
    for i in queryset:
        i.revoke()


reexecute_tasks.short_description = "Taskを再実行する"
cancel_revoke_tasks.short_description = "Taskをcancel/revokeする"


@admin.register(models.CeleryTask)
class CeleryTaskAdmin(admin.ModelAdmin):
    model = models.CeleryTask
    list_display = ('id', 'status', 'user', 'task_id', 'task_path',
                    'created_at', 'updated_at', 'start_at', 'end_at')
    list_filter = ('status', )

    actions = [reexecute_tasks, cancel_revoke_tasks, ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from future.utils import python_2_unicode_compatible
import yaml

from celery.task.control import revoke as celery_revoke

from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string
from django.utils.functional import SimpleLazyObject

from . import utils


TASK_STATE = (
    # from celery.events.state
    ('sent', 'PENDING', ),
    ('received', 'RECEIVED', ),
    ('started', 'STARTED', ),
    ('failed', 'FAILURE', ),
    ('retried', 'RETRY', ),
    ('succeeded', 'SUCCESS', ),
    ('revoked', 'REVOKED', ),
    ('rejected', 'REJECTED', ),

    # custom state
    ('cancel', 'CANCEL'),
    ('cancelled', 'CANCELLED'),
    ('revoking', 'REVOKING'),
    ('retry_rejected', 'RETRY_REJECTED'),
)


@python_2_unicode_compatible
class Worker(models.Model):
    worker_id = models.CharField('ワーカーID', max_length=256)
    started_at = models.DateTimeField(auto_now_add=True)
    ended_at = models.DateTimeField(blank=True, null=True)
    beated_at = models.DateTimeField(blank=True, null=True)

    class State:
        RUNNING = 0
        STOPPED = 1
        BEAT_FAILED = 2

    STATE_CHOICES = (
        (State.RUNNING, "開始"),
        (State.STOPPED, "停止"),
        (State.BEAT_FAILED, "応答なし"),
    )

    status = models.IntegerField(choices=STATE_CHOICES, default=State.RUNNING)

    def __str__(self):
        return 'worker({})'.format(self.worker_id)


@python_2_unicode_compatible
class WorkerTaskLog(models.Model):
    worker = models.ForeignKey(Worker)
    task_id = models.CharField(max_length=256, unique=True)
    task_path = models.CharField(max_length=256)
    started_at = models.DateTimeField(auto_now_add=True)
    ended_at = models.DateTimeField(blank=True, null=True)


@python_2_unicode_compatible
class TaskSignature(models.Model):
    name = models.CharField('名前', max_length=256, blank=True, null=True)
    description = models.TextField('詳細', blank=True, null=True)

    task_path = models.CharField('タスク名', max_length=256)
    args = models.TextField('引数', blank=True, null=True)
    kwargs = models.TextField('kw引数', blank=True, null=True)
    options = models.TextField('タスクオプション', blank=True, null=True)

    def _read_yaml(self, value, default_value):
        if not value:
            return default_value
        return yaml.load(value)

    def get_args(self):
        return self._read_yaml(self.args, ())

    def get_kwargs(self):
        return self._read_yaml(self.kwargs, {})

    def get_options(self):
        return self._read_yaml(self.options, {})

    def run(self):
        import_string(self.task_path).apply_async(
            args=self.get_args(), kwargs=self.get_kwargs(), **self.get_options())

    def save(self, *args, **kwargs):
        if not (isinstance(self.get_args(), (tuple, list)) and
                isinstance(self.get_kwargs(), dict) and
                isinstance(self.get_options(), dict)):
            raise TypeError()
        return super(TaskSignature, self).save(*args, **kwargs)

    def __str__(self):
        return self.name if self.name else self.task_path


@python_2_unicode_compatible
class CronSchedule(models.Model):
    name = models.CharField('名前', max_length=256, blank=True, null=True)
    description = models.TextField('詳細', blank=True, null=True)

    cron_expr = models.CharField('cron', max_length=256)
    next_time = models.DateTimeField(blank=True, null=True)
    task = models.ForeignKey(TaskSignature)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    expires = models.DateTimeField('有効期限', blank=True, null=True)

    last_run_at = models.DateTimeField('最終実行日時', blank=True, null=True)
    total_run_count = models.PositiveIntegerField('実行回数', default=0, editable=False)
    max_run_count = models.IntegerField('最大繰り返し回数', default=-1)

    def run(self, now):
        self.task.run()
        self.last_run_at = utils.get_now()
        self.total_run_count += 1
        self.next_time = utils.get_next_cron(self.cron_expr)
        self.save()

    def save(self, *args, **kwargs):
        update_next_time = not self.next_time
        if self.pk and hasattr(self, 'cron_expr'):
            old = self.__class__.objects.get(pk=self.pk)
            update_next_time = old.cron_expr != self.cron_expr
        if update_next_time:
            self.next_time = utils.get_next_cron(self.cron_expr)
        return super(CronSchedule, self).save(*args, **kwargs)

    def __str__(self):
        return '{} {} {}'.format(
            self.name if self.name else '',
            self.cron_expr, self.task)


class CeleryTask(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True)
    task_id = models.CharField('タスクID', max_length=256, unique=True)

    task_path = models.CharField('タスク名', max_length=256, blank=True, null=True)
    args = models.TextField('引数', blank=True, null=True)
    kwargs = models.TextField('kw引数', blank=True, null=True)

    status = models.CharField(
        'ステータス', max_length=64, choices=TASK_STATE, default='sent')
    stack_trace = models.TextField('スタックトレース', blank=True, null=True)
    result_string = models.TextField('結果', blank=True, null=True)
    result_file = models.FileField(
        '結果ファイル', storage=SimpleLazyObject(import_string(
            settings.CELERY_TASK_FILE_STORAGE if hasattr(settings, 'CELERY_TASK_FILE_STORAGE')
            else settings.DEFAULT_FILE_STORAGE)), blank=True, null=True)

    created_at = models.DateTimeField(auto_now_add=True)
    start_at = models.DateTimeField(editable=False, blank=True, null=True)
    end_at = models.DateTimeField(editable=False, blank=True, null=True)
    updated_at = models.DateTimeField(auto_now=True)

    def reexecute(self, **options):
        import_string(self.task_path).apply_async(
            args=yaml.load(self.args), kwargs=yaml.load(self.kwargs), **options)

    def revoke(self):
        if self.status == 'sent':
            self.status = 'cancel'
        else:
            self.status = 'revoking'
        celery_revoke(self.task_id, terminate=True)
        self.save()
